(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Graph imports ErrorMonad Show begin

types
 'a graph = "'a set \<times> ('a \<times> 'a) set"
 'a ipath = "nat \<Rightarrow> 'a"
 'a fpath = "'a list"

definition nodes :: "'a graph \<Rightarrow> 'a set"
where "nodes G = fst G"

definition edges :: "'a graph \<Rightarrow> ('a \<times> 'a) set"
where "edges G = snd G"

inductive path :: "'a graph \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> bool"
where refl: "path G n n"
    | trans: "(l,m) \<in> edges G \<Longrightarrow> path G m n \<Longrightarrow> path G l n"

fun is_fin_path :: "'a graph \<Rightarrow> 'a fpath \<Rightarrow> bool"
where "is_fin_path G [] = True"
    | "is_fin_path G [x] = True"
    | "is_fin_path G (x # y # xs) = ((x,y) \<in> edges G \<and> is_fin_path G (y # xs))"

lemmas is_fin_path_induct[case_names baseI baseII step] = is_fin_path.induct

lemma fin_path_imp_path: assumes "is_fin_path G (x#xs)"
      shows "path G x (last(x#xs))"
using assms proof (induct xs arbitrary: x)
 case Nil thus ?case using refl by simp
next
 case (Cons y ys)
 have A:"last(x#y#ys) = last(y#ys)" by (induct ys) auto
 from Cons have "is_fin_path G (y#ys)" by simp
 with Cons have "(x,y) \<in> edges G" and "path G y (last(y#ys))" by auto
 thus ?case unfolding A using path.trans[where ?G = G] by simp
qed
 
lemma path_exists: assumes A: "path G i j"
      shows "\<exists> xs. last (i # xs) = j \<and> is_fin_path G (i # xs)" (is "?P i")
      using `path G i j`
proof (induct)
  case (refl G i)
    have "last [i] = i" and "is_fin_path G [i]" by auto
    thus ?case by auto
  next
  case (trans l m G j)
    from `\<exists>xs. last (m # xs) = j \<and> is_fin_path G (m # xs)` obtain xs 
     where "last (m # xs) = j \<and> is_fin_path G (m # xs)" ..
    from this have B: "last (m # xs) = j" and C: "is_fin_path G (m # xs)" by auto 
    from C and `(l,m) \<in> edges G` have D: "is_fin_path G (l # m # xs)" by simp
    from B have "last (l # m # xs) = j" by simp
    from this and D show ?case by blast
qed


definition is_path :: "'a graph \<Rightarrow> 'a ipath \<Rightarrow> bool"
where "is_path G f = (\<forall> n. (f n, f (Suc n)) \<in> edges G)"


definition node_restriction :: "'a graph \<Rightarrow> 'a set \<Rightarrow> 'a graph" ("_|\<^sub>\<^sub>\<^sub>_" 60)
where "G|\<^sub>\<^sub>\<^sub>N = (nodes G \<inter> N, {(m,n). m \<in> N \<and> n \<in> N \<and> (m,n) \<in> edges G})"

definition cycle :: "'a set \<Rightarrow> 'a graph \<Rightarrow> bool"
where
 "cycle C G = (C \<noteq> {} \<and> (\<forall>m \<in> C. \<forall> n \<in> C. path (G|\<^sub>\<^sub>\<^sub>C) m n))"

definition fully_connected :: "'a set \<Rightarrow> 'a graph" 
  where "fully_connected S = (S, {(m,n). m \<in> S \<and> n \<in> S})"

lemma is_fin_path_restriction[rule_format] : "is_fin_path (G|\<^sub>\<^sub>\<^sub>C) xs \<longrightarrow> is_fin_path G xs"
  by (induct rule:is_fin_path.induct[where ?P = "\<lambda> G xs. is_fin_path (G|\<^sub>\<^sub>\<^sub>C) xs \<longrightarrow> is_fin_path G xs"], 
    unfold node_restriction_def edges_def, auto, unfold edges_def, auto)

lemma fin_path_subset_imp_fin_path: "is_fin_path (G|\<^sub>\<^sub>\<^sub>C) xs \<Longrightarrow> C \<subseteq> D \<Longrightarrow> is_fin_path (G|\<^sub>\<^sub>\<^sub>D) xs"
proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons y ys)
 hence IH: "is_fin_path(G|\<^sub>\<^sub>\<^sub>C) ys \<Longrightarrow> is_fin_path (G|\<^sub>\<^sub>\<^sub>D) ys" by simp
 show ?case
 proof (cases ys)
  case Nil thus ?thesis by simp
 next
  case (Cons z zs)
  from `is_fin_path (G|\<^sub>\<^sub>\<^sub>C) (y#ys)`
  have "is_fin_path (G|\<^sub>\<^sub>\<^sub>C) (z#zs)" and "(y,z) \<in> edges(G|\<^sub>\<^sub>\<^sub>C)"
   unfolding `ys=z#zs`  by (induct ys) auto
  with IH have "is_fin_path (G|\<^sub>\<^sub>\<^sub>D) ys" unfolding `ys=z#zs` by simp
  from `C \<subseteq> D` and `(y,z) \<in> edges(G|\<^sub>\<^sub>\<^sub>C)` have "(y,z) \<in> edges(G|\<^sub>\<^sub>\<^sub>D)"
   unfolding edges_def node_restriction_def by auto
  with `is_fin_path (G|\<^sub>\<^sub>\<^sub>D) ys` show ?thesis unfolding `ys=z#zs` by simp
 qed
qed

lemma path_monotone: assumes "path (G|\<^sub>\<^sub>\<^sub>C) m n" and "C \<subseteq> D" shows "path (G|\<^sub>\<^sub>\<^sub>D) m n"
proof -
 from `path (G|\<^sub>\<^sub>\<^sub>C) m n` obtain xs where "is_fin_path (G|\<^sub>\<^sub>\<^sub>C) (m#xs)" and "n = last(m#xs)"
  using path_exists by best
 with fin_path_subset_imp_fin_path and `C \<subseteq> D` have "is_fin_path (G|\<^sub>\<^sub>\<^sub>D) (m#xs)" by simp
 thus ?thesis using fin_path_imp_path[where G = "G|\<^sub>\<^sub>\<^sub>D" and x = m]
  unfolding `n = last(m#xs)` by simp
qed

lemma is_path_monotone: assumes "is_path (G|\<^sub>\<^sub>\<^sub>C) p" and "C \<subseteq> D" shows "is_path (G|\<^sub>\<^sub>\<^sub>D) p" 
using assms by (unfold is_path_def node_restriction_def edges_def, auto)

lemma is_fin_path_trans: "y = last (x # xs) \<Longrightarrow> is_fin_path G (x # (xs @ ys)) = (is_fin_path G (x # xs) \<and> is_fin_path G (y # ys))"
by (induct xs arbitrary: x, simp_all)


lemma path_trans[trans]: assumes "path G m n" and "path G n k" shows "path G m k" 
proof -
  from assms obtain xs where mn1: "is_fin_path G (m#xs)" and mn2: "n = last (m#xs)" using path_exists by best
  from assms obtain ys where nk1: "is_fin_path G (n#ys)" and nk2: "k = last (n#ys)" using path_exists by best
  from mn1 and mn2 and nk1 have "is_fin_path G (m # (xs @ ys))" by (simp add: is_fin_path_trans)
  hence "path G m (last (m # (xs @ ys)))" by (rule fin_path_imp_path)
  from this and mn2 and nk2 show ?thesis by auto
qed
  

definition scc_of_node :: "'a graph \<Rightarrow> 'a \<Rightarrow> 'a set"
where "scc_of_node G n = (\<Union> {C. (n \<in> C \<and> cycle C G)})"

lemma node_in_scc_of_node:
  "n \<in> scc_of_node G n"
proof -
  have "cycle {n} G" by (unfold cycle_def, simp_all add: path.refl)
  thus ?thesis by (unfold scc_of_node_def, auto)
qed
  

lemma cycle_Union:
  "cycle (\<Union> {C. (k \<in> C \<and> cycle C G)}) G" (is "cycle ?U G")
proof -
  have "k \<in> scc_of_node G k" by (rule node_in_scc_of_node)
  hence NE: "?U \<noteq> {}" by (unfold scc_of_node_def, auto)
  have "\<forall> m \<in> ?U. \<forall> n \<in> ?U. path (G|\<^sub>\<^sub>\<^sub>?U) m n"
  proof 
    fix m
    assume "m \<in> ?U"
    from this obtain D where "m \<in> D" and "k \<in> D" and "cycle D G" and DU: "D \<subseteq> ?U" by auto
    hence "(path (G|\<^sub>\<^sub>\<^sub>D) m k)" by (unfold cycle_def, simp_all)
    from this and DU have mk: "path (G|\<^sub>\<^sub>\<^sub>?U) m k" by (rule path_monotone)
    show "\<forall> n \<in> ?U. path (G|\<^sub>\<^sub>\<^sub>?U) m n"
    proof 
      fix n
      assume "n \<in> ?U"
      from this obtain E where "n \<in> E" and "k \<in> E" and "cycle E G" and EU: "E \<subseteq> ?U" by auto
      hence "(path (G|\<^sub>\<^sub>\<^sub>E) k n)" by (unfold cycle_def, simp_all)
      from this and EU have "path (G|\<^sub>\<^sub>\<^sub>?U) k n" by (rule path_monotone)
      from mk and this show "path (G|\<^sub>\<^sub>\<^sub>?U) m n" by (rule path_trans)
   qed
  qed
  from NE and this show ?thesis by (unfold cycle_def, simp_all)
qed
    
lemma cycle_union: 
      assumes cC: "cycle C G" and
              cD: "cycle D G" and
              "C \<inter> D \<noteq> {}"
      shows "cycle (C \<union> D) G"
      using assms
proof (unfold cycle_def, simp_all)
  from cC have pC: "\<forall> m \<in> C. \<forall> n \<in> C.(path (G|\<^sub>\<^sub>\<^sub>C) m n)" by (unfold cycle_def,auto)
  from cD have pD: "\<forall> m \<in> D. \<forall> n \<in> D.(path (G|\<^sub>\<^sub>\<^sub>D) m n)" by (unfold cycle_def,auto)
  from `C \<inter> D \<noteq> {}` obtain x where xC: "x \<in> C" and xD: "x \<in> D" by auto
  have CCD: "C \<subseteq> C \<union> D" by auto
  have DCD: "D \<subseteq> C \<union> D" by auto
  show "\<forall> m \<in> C \<union> D. \<forall> n \<in> C \<union> D. path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
  proof 
    fix m
    assume "m \<in> C \<union> D"
    thus "\<forall> n \<in> C \<union> D. path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
    proof 
      assume "m \<in> C"
      show "\<forall> n \<in> C \<union> D. path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
      proof
	fix n
	assume "n \<in> C \<union> D"
	thus "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
	proof
	  next
	  assume "n \<in> C"
	  from this and `m \<in> C` and pC have "path (G|\<^sub>\<^sub>\<^sub>C) m n" by simp
	  from this and CCD show ?thesis by (rule path_monotone)
          next
	  assume "n \<in> D"
	  from this and xD and pD have "path (G|\<^sub>\<^sub>\<^sub>D) x n" by simp
	  from this and DCD have xn: "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) x n" by (rule path_monotone)
	  from `m \<in> C` and xC and pC have "path (G|\<^sub>\<^sub>\<^sub>C) m x" by simp
	  from this and CCD have mx: "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m x" by (rule path_monotone)
	  from this and xn show ?thesis by (rule path_trans)
        qed
      qed   
    next
      assume "m \<in> D"
      show "\<forall> n \<in> C \<union> D. path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
      proof
	fix n
	assume "n \<in> C \<union> D"
	thus "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m n"
	proof
	  next
	  assume "n \<in> D"
	  from this and `m \<in> D` and pD have "path (G|\<^sub>\<^sub>\<^sub>D) m n" by simp
	  from this and DCD show ?thesis by (rule path_monotone)
          next
	  assume "n \<in> C"
	  from this and xC and pC have "path (G|\<^sub>\<^sub>\<^sub>C) x n" by simp
	  from this and CCD have xn: "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) x n" by (rule path_monotone)
	  from `m \<in> D` and xD and pD have "path (G|\<^sub>\<^sub>\<^sub>D) m x" by simp
	  from this and DCD have mx: "path (G|\<^sub>\<^sub>\<^sub>C \<union> D) m x" by (rule path_monotone)
	  from this and xn show ?thesis by (rule path_trans)
        qed
      qed   
    qed
  qed
qed


definition scc :: "'a set \<Rightarrow> 'a graph \<Rightarrow> bool"
where
 "scc C G = (cycle C G \<and> \<not> (\<exists> D. cycle D G \<and> C \<subset> D))"

lemma scc_Union : "scc (\<Union> {C. (k \<in> C \<and> cycle C G)}) G" (is "scc ?U G")
proof -
  have CU: "cycle ?U G" by (rule cycle_Union) 
  have "\<not> (\<exists> D. cycle D G \<and> ?U \<subset> D)"
  proof 
    assume "(\<exists> D. cycle D G \<and> ?U \<subset> D)"
    from this obtain D where cD: "cycle D G" and UD: "?U \<subset> D" by auto
    have "k \<in> scc_of_node G k" by (rule node_in_scc_of_node)
    from this and UD have "k \<in> D" by (unfold scc_of_node_def, auto)
    from this and cD have "D \<subseteq> ?U" by auto
    from this and UD show False by auto
  qed
  from this and CU show ?thesis by (unfold scc_def, simp)
qed

lemma scc_of_node_is_scc : "scc (scc_of_node G n) G"
by (unfold scc_of_node_def, simp add: scc_Union)

lemma scc_disjoint: 
  assumes C: "scc C G" 
      and D: "scc D G"
      and NE: "C \<noteq> D"
    shows "C \<inter> D = {}"
proof (rule ccontr)
  assume NEmpty: "C \<inter> D \<noteq> {}"
  from NE obtain x where "(x \<in> C) \<noteq> (x \<in> D)" by blast
  hence "\<exists> E. \<exists> F. (scc E G) \<and> (scc F G) \<and> x \<in> E \<and> x \<notin> F \<and> E \<inter> F \<noteq> {}" 
  proof (cases "x \<in> C")
    assume "x \<in> C"
    from this and C and D and `(x \<in> C) \<noteq> (x \<in> D)` and NEmpty show ?thesis by blast 
  next
    assume nc: "x \<notin> C"
    from this and C and D and `(x \<in> C) \<noteq> (x \<in> D)` and NEmpty show ?thesis by blast 
  qed
  from this obtain E and F where 
       E: "(scc E G)" 
   and F: "(scc F G)"
   and xE: "x \<in> E"
   and xF: "x \<notin> F" 
   and NEI: "E \<inter> F \<noteq> {}" by auto
  show False
  proof -
    from NEI and E and F have cEF: "cycle (E \<union> F) G" by (unfold scc_def, simp add: cycle_union)
    from xE and xF have "F \<subset> E \<union> F" by auto
    from F and this cEF show False by (unfold scc_def, simp) 
  qed
qed


             
definition sccs :: "'a graph \<Rightarrow> ('a set) set"
where 
  "sccs G = { C. scc C G}"


definition condense_graph :: "'a graph \<Rightarrow> ('a set) graph"
where 
  "condense_graph G = (sccs G, {(C,D). scc C G \<and> scc D G \<and> (\<exists> n m. n \<in> C \<and> m \<in> D \<and> (n,m) \<in> edges G) })"
  

definition condense_path :: "'a graph \<Rightarrow> 'a ipath \<Rightarrow> ('a set) ipath"
where 
  "condense_path G p = (\<lambda> i. scc_of_node G (p i))"

lemma condense_sound : assumes "is_path G p" shows "is_path (condense_graph G) (condense_path G p)" (is "is_path ?CG ?CP") using assms
proof -
  assume "is_path G p"
  hence A: "\<forall> n. (p n, p (Suc n)) \<in> edges G" by (unfold is_path_def)
  have "\<forall> n. (?CP n, ?CP (Suc n)) \<in> edges ?CG"
  proof 
    fix n
    from A have E: "(p n, p (Suc n)) \<in> edges G" by simp
    have "p n \<in> scc_of_node G (p n)" (is "_ \<in> ?spn")
     and "p (Suc n) \<in> scc_of_node G (p (Suc n))"  (is "_ \<in> ?spsn")
      by (simp_all add:node_in_scc_of_node)
    from this and E have newEdge: "(\<exists>n. n \<in> ?spn \<and> (\<exists>m. m \<in> ?spsn \<and> (n, m) \<in> edges G))" by auto
    have nscc: "scc ?spn G"
     and snscc: "scc ?spsn G"
      by (simp_all add:scc_of_node_is_scc)
    from this and newEdge show "((condense_path G p) n, (condense_path G p) (Suc n)) \<in> edges (condense_graph G)"
     by (unfold condense_graph_def condense_path_def edges_def, auto)
  qed
  thus ?thesis by (unfold is_path_def, simp)
qed

definition path_edges :: "'a ipath \<Rightarrow> ('a \<times> 'a) set" 
where "path_edges p = ({(p n, p (Suc n)) | n . True})"

definition path_nodes :: "'a ipath \<Rightarrow> 'a set"
where "path_nodes p = range p"


definition suffix :: "'a ipath \<Rightarrow> 'a ipath \<Rightarrow> bool" ("_ =\<dots> _" 50)
where "(p =\<dots> q) = (\<exists> n. \<forall> m. (p (m+n) = q m))"
  

lemma suffix_trans[trans] : assumes "p =\<dots> q" and "q =\<dots> r" shows "p =\<dots> r" using assms
proof (unfold suffix_def)
  assume "\<exists>n. \<forall>m. p (m + n) = q m" 
  from this obtain n1 where A: "\<forall> m. p (m + n1) = q m" ..
  assume "\<exists>n. \<forall>m. q (m + n) = r m"
  from this obtain n2 where B: "\<forall>m. q (m + n2) = r m" ..
  have " \<forall>m. p (m + (n2 + n1)) = (r m)" 
  proof 
    fix m
    have "m + (n2 + n1) = m + n2 + n1" (is "?x = ?y") by arith
    hence "p (m + (n2 + n1)) = p (m + n2 + n1)" using arg_cong[where x = ?x and y = ?y] by simp
    also from A have "\<dots> = q (m + n2) " ..
    also from B have "\<dots> = r m" ..
    finally show "p (m + (n2 + n1)) = r m" .
 qed
 thus  "\<exists>n. \<forall>m. p (m + n) = r m" ..
qed
  

definition drop_path :: "'a ipath \<Rightarrow> nat \<Rightarrow> 'a ipath"
where "drop_path p n = (\<lambda> m. p (m + n))"

lemma is_path_imp_suffix_is_path: assumes "is_path G p" and "p =\<dots> q" shows "is_path G q" 
proof -
  from `p =\<dots> q` obtain k where A:"\<forall> m. p (m + k) = q m" by (unfold suffix_def, auto)
  from `is_path G p` have B: "(\<forall> n. (p n, p (Suc n)) \<in> edges G)" by (unfold is_path_def)
  have "\<forall> n. (q n, q (Suc n)) \<in> edges G"
  proof 
    fix n
    have "Suc (n+k) = (Suc n) + k" by simp
    from A have C: "q n = p (n+k)" by simp 
    from A have D: "p ((Suc n)+k) = q (Suc n) " .. 
    from B have "(p (n+k), p (Suc (n+k))) \<in> edges G" ..
    from this and C and D show "(q n, q (Suc n)) \<in> edges G" by simp
  qed
  thus ?thesis by (unfold is_path_def)
qed

lemma drop_path_suffix: "p =\<dots> drop_path p n"
by (unfold drop_path_def suffix_def, auto)

lemma drop_path_is_path: "is_path G p \<Longrightarrow> is_path G (drop_path p n)"
by (simp add: is_path_imp_suffix_is_path drop_path_suffix)

lemma path_edges_monotone: assumes "p =\<dots> q" shows "path_edges p \<supseteq> path_edges q" using assms
proof (unfold path_edges_def suffix_def) 
  assume "\<exists>n. \<forall>m. p (m + n) = q m"
  from this obtain n where A: " \<forall>m. p (m + n) = q m" .. 
  show "{(q n, q (Suc n)) |n. True} \<subseteq> {(p n, p (Suc n)) |n. True}"
  proof 
    fix x
    assume "x \<in> {(q n, q (Suc n)) | n. True}" 
    from this obtain m where X: "x = (q m, q (Suc m))" by auto
    from A have B: "p (m+n) = q m" ..
    have "p (Suc (m+n)) = p ((Suc m) + n)" by simp 
    also from A have "\<dots> = q (Suc m)" ..
    finally have "x = (p (m+n), p ((Suc m) + n))" using B and X by simp
    thus " x \<in> {(p n, p (Suc n)) |n. True}" by auto
  qed
qed

lemma is_path_imp_path : assumes "is_path G p" and  "m = n + k"
  shows "path G (p n) (p m)" using assms
proof (induct k arbitrary: n)
 case 0
   have A: "path G (p n) (p n)" by (rule path.refl)
   from `m = n + 0` have "m = n" by simp
   from this and A show ?case by simp
 next
 case (Suc k)
   hence B: "path G (p (Suc n)) (p m)" by simp
   from `is_path G p` have "(p n, p (Suc n)) \<in> edges G" by (unfold is_path_def, simp)
   from this and B show "path G (p n) (p m)" by (rule path.trans)
qed

lemma is_path_imp_path_2 : assumes "is_path G p" and "m \<ge> n"
  shows "path G (p n) (p m)" 
proof -
  from assms have "\<exists> k. m = n + k" by arith
  from this obtain k where "m = n + k" .. 
  from assms and this show ?thesis by (simp add: is_path_imp_path)
qed


lemma edges_contained : assumes "j < length xs - 1" and "is_fin_path G xs" 
  shows "(xs ! j, xs ! (Suc j)) \<in> edges G" (is "?E G xs j") 
proof -
  have "\<forall> j. is_fin_path G xs \<and> j < length xs - 1 \<longrightarrow> (xs ! j, xs ! (Suc j)) \<in> edges G"
  proof (induct rule: is_fin_path_induct[where ?P = "\<lambda> G xs. (\<forall> j. (is_fin_path G xs \<and> j < length xs - 1 \<longrightarrow> (xs ! j, xs ! (Suc j)) \<in> edges G))"])
    case (baseI G) thus ?case by simp
  next
    case (baseII G x) thus ?case by simp
  next
    case (step G x y xs) show ?case 
    proof 
      fix j
      from step show "is_fin_path G (x # y # xs) \<and> j < length (x # y # xs) - 1 \<longrightarrow>
        ((x # y # xs) ! j, (x # y # xs) ! Suc j) \<in> edges G" by (cases j, auto)
    qed
  qed
  with assms show ?thesis by blast
qed


lemma partial_path : assumes "j = i + k" 
    and "j < length xs" 
    and A: "is_fin_path G xs" 
    shows "path (G|\<^sub>\<^sub>\<^sub>(set xs)) (xs ! i) (xs ! j)" using assms
proof (induct k arbitrary: i)
  case 0
    from `j = i + 0` have "i = j" by simp
    hence "xs ! i  = xs ! j" by simp
    thus ?case by (simp add: path.refl)
  next
  case (Suc k)
    hence B: "path (G|\<^sub>\<^sub>\<^sub>(set xs)) (xs ! (Suc i)) (xs ! j)"  by simp
    assume "j = i + Suc k" and "j < length xs" hence "i < length xs - 1" by simp
    from this and A have edge: "(xs ! i, xs ! (Suc i)) \<in> edges G" by (rule edges_contained)
    from `j = i + Suc k` and `j < length xs` have "xs ! i \<in> set xs" and "xs ! (Suc i) \<in> set xs" by auto
    with edge have "(xs ! i, xs ! (Suc i)) \<in> edges (G|\<^sub>\<^sub>\<^sub>(set xs))" by (unfold node_restriction_def edges_def, simp)
    from this and B show "path (G|\<^sub>\<^sub>\<^sub>(set xs)) (xs ! i) (xs ! j)" by (rule path.trans)
qed    

lemma list_set : assumes "x \<in> set xs" shows "\<exists> i . (xs ! i = x) \<and> (length xs > i)" using assms
proof (induct xs)
 case Nil thus ?case by simp next
 case (Cons y ys)
   thus ?case 
   proof (cases "x = y")
     case True
       hence "(y # ys) ! 0 = x \<and> (length (y # ys) > 0)" by simp thus ?thesis ..
     next
     case False
       from this and `x \<in> set (y # ys)` have "x \<in> set ys" by simp
       from this and `x \<in> set ys \<Longrightarrow> \<exists>i. ys ! i = x \<and> i < length ys` have "\<exists> i. ys ! i = x \<and> i < length ys" by simp
       from this obtain i where "ys ! i = x \<and> i < length ys" ..
       hence "(y # ys) ! (Suc i) = x \<and> (Suc i) < length (y # ys)" by simp
       thus ?thesis ..
   qed
qed
  
lemma last_nth : assumes "xs \<noteq> []" shows "last xs = xs ! (length xs - 1)" using assms 
proof (induct xs)
  case Nil thus ?case by simp
next
  case (Cons y ys)
  thus ?case
  proof (cases ys)
    case Nil
      thus ?thesis by simp
    next 
    case (Cons z zs)
      from this and `ys \<noteq> [] \<Longrightarrow> last ys = ys ! (length ys - 1)` have "last ys = ys ! (length ys - 1)" by simp
      from this and `ys = z # zs` show "last (y # ys) = (y # ys) ! (length (y # ys) - 1)" by (simp)
  qed
qed

lemma cycle_cycle : assumes fP: "is_fin_path G (x#xs)" 
                        and xl: "last (x#xs) = x" 
                      shows "cycle (set (x#xs)) G"
proof (unfold cycle_def)
  have easy: "set (x # xs) \<noteq> {}" (is ?PartI) by simp
  have hard: "\<forall>m\<in>set (x # xs). \<forall> n \<in> set (x # xs). path (G|\<^sub>\<^sub>\<^sub>set (x # xs)) m n" (is ?PartII)
  proof
    fix m 
    let ?GS = "G|\<^sub>\<^sub>\<^sub>(set (x # xs))"
    assume mS: "m \<in> set (x # xs)" show "\<forall> n \<in> set (x # xs). path ?GS m n"
    proof 
      fix n
      assume nS: "n \<in> set (x # xs)"
      show "path ?GS m n"
      proof -
        from mS have "\<exists> i. ((x # xs) ! i = m \<and> i < length (x # xs))" by (rule list_set)
	from this obtain i where mi: "(x # xs) ! i = m" and ik: "i < length (x # xs)" by auto
        from nS have "\<exists> j. ((x # xs) ! j = n \<and> j < length (x # xs))" by (rule list_set)
	from this obtain j where nj: "(x # xs) ! j = n" and jk: "j < length (x # xs)" by auto
	have "j = 0 + j" by simp
	from this and jk and fP have "path ?GS ((x # xs) ! 0) ((x # xs) ! j)" by (rule partial_path)
	from this and nj have latterPath: "path ?GS x n" by simp
	let ?k = "length (x # xs) - 1"
	have "x # xs \<noteq> []" by simp hence lastK: "last (x # xs) = (x # xs) ! ?k" by (rule last_nth)
	from ik have "\<exists> l. ?k = i + l" by arith
	from this obtain l where lk: "?k = i + l" ..
	have "?k < length (x # xs)" by simp
        from lk and this and fP have "path ?GS ((x # xs) ! i) ((x # xs) ! ?k)" by (rule partial_path)
	from this and mi and lastK and xl have "path ?GS m x" by simp
	from this and latterPath show "path ?GS m n" by (rule path_trans)
      qed
    qed
  qed
  from easy and hard show "?PartI \<and> ?PartII" ..
qed



lemma cycle_scc : assumes "path G n m" and "path G m n" shows "m \<in> scc_of_node G n" using assms
proof -
  from `path G n m` have "\<exists> xs. last (n # xs) = m \<and> is_fin_path G (n#xs)" by (rule path_exists)
  from this obtain ns where nm: "last (n # ns) = m" and pnm: "is_fin_path G (n # ns)" by auto
  from `path G m n` have "\<exists> xs. last (m # xs) = n \<and> is_fin_path G (m#xs)" by (rule path_exists)
  from this obtain ms where mn: "last (m # ms) = n" and pmn: "is_fin_path G (m # ms)" by auto
  from nm and pnm and pmn have path: "is_fin_path G (n # ns @ ms)" by  (simp add: is_fin_path_trans)
  from nm and mn have "last (n # ns @ ms) = n"
  proof (cases "ns = []")
    case True
      from this and nm and mn show ?thesis by simp
    next
    case False
      assume A: "ns \<noteq> []"
      from this and nm and mn show ?thesis 
      proof (cases "ms = []")
	case True
	from this and nm and mn and A show ?thesis by simp
      next
	case False
	from this and nm and mn and A show ?thesis by simp
      qed
  qed
  from path and this have cyc: "cycle (set (n # ns @ ms)) G" (is "cycle ?C G") by (rule cycle_cycle)
  have nC: "n \<in> ?C" by simp
  from nm have mC: "m \<in> ?C" by auto
  from cyc and nC have "?C \<subseteq> scc_of_node G n" by (unfold scc_of_node_def, auto) 
  from mC and this show ?thesis by auto
qed


lemma main_scc_helper: assumes "finite E"
        and "is_path G p"
        and "path_edges p \<subseteq> E"
      shows "(\<exists> q. \<exists> C. (scc C G \<and> p =\<dots> q \<and> (path_nodes q \<subseteq> C)))" 
using assms
proof (induct E arbitrary: p)
  case empty
    from `path_edges p \<subseteq> {}` show ?case
     by (unfold path_edges_def, blast)
  next
  case (insert x F)
    thus ?case
      proof (cases "\<exists> n. (path_edges (drop_path p n) \<subseteq> F)")
	case True
	from this obtain n where A: "path_edges (drop_path p n) \<subseteq> F" by auto
	from `is_path G p` have B: "is_path G (drop_path p n)" by (rule drop_path_is_path)	
	have "p =\<dots> drop_path p n" by (rule drop_path_suffix)
	from `\<And>p. \<lbrakk>is_path G p; path_edges p \<subseteq> F\<rbrakk> \<Longrightarrow> \<exists>q C. scc C G \<and> p =\<dots> q \<and> path_nodes q \<subseteq> C`
	  and A and B have "\<exists>q C. scc C G \<and> (drop_path p n) =\<dots> q \<and> path_nodes q \<subseteq> C" by simp
	from this obtain q C where D: "scc C G \<and> (drop_path p n) =\<dots> q \<and> path_nodes q \<subseteq> C" by auto
	hence "drop_path p n =\<dots> q" by simp
	from `p =\<dots> drop_path p n` and this have "p =\<dots> q" by (rule suffix_trans)
	from this and D have "scc C G \<and> p =\<dots> q \<and> path_nodes q \<subseteq> C" by simp
	thus ?thesis by auto
      next 
      case False
        assume pPath: "is_path G p"
	assume noN: " \<not> (\<exists>n. path_edges (drop_path p n) \<subseteq> F)" 
	  and  pxF: "path_edges p \<subseteq> insert x F"
	hence B: "\<forall> m. \<exists> k \<ge> m. x = ((p k, p (Suc k)))"	
	proof -
	  show ?thesis
	  proof 
	    fix m
	    from noN have "\<not> (path_edges (drop_path p m) \<subseteq> F)" by simp
	    have "p =\<dots> drop_path p m" by (rule drop_path_suffix) 
	    hence "path_edges p \<supseteq> path_edges (drop_path p m)" by (rule path_edges_monotone)
	    from this and pxF have "path_edges (drop_path p m) \<subseteq> insert x F"  by simp
	    from this and noN have "x \<in> path_edges (drop_path p m)" by auto
	    hence "\<exists> l. x = ((drop_path p m) l, (drop_path p m) (Suc l))" 
	      by (unfold drop_path_def path_edges_def, simp)
	    from this obtain l where "x = ((drop_path p m) l, (drop_path p m) (Suc l))" ..
	    hence "x = (p (l+m), p ((Suc l) + m))" by (unfold drop_path_def, auto)
	    hence nearly: "x = (p (l+m), p (Suc (l+m))) " by simp
	    show "\<exists>k\<ge>m. x = (p k, p (Suc k))" 
	      proof
		have "l + m \<ge> m" by simp
		from this and nearly show "m \<le> l+m \<and> x = (p (l+m), p (Suc (l+m)))" by simp
	      qed
	 qed
      qed
      hence "\<exists>k\<ge>0. x = (p k, p (Suc k))" ..
      from this obtain k where xpk: "x = (p k, p (Suc k))" by auto
      let ?pk = "p k"
      let ?C = "scc_of_node G ?pk" 
      let ?q = "drop_path p k"
      have C: "p =\<dots> ?q" by (rule drop_path_suffix)
      have D: "scc ?C G" by (rule scc_of_node_is_scc)
      from pPath have qPath: "is_path G ?q" by (rule drop_path_is_path)
      have "path_nodes ?q \<subseteq> ?C" 
      proof (unfold path_nodes_def, rule subsetI)
	fix y
	assume "y \<in> range ?q"
	from this obtain l where ly: "?q l = y" by auto
	have zpk: "(drop_path p k) 0 = ?pk" by (unfold drop_path_def, simp)
	have "0 \<le> l" by simp
	from qPath and this  have "path G (?q 0) (?q l)" by (rule is_path_imp_path_2)
	from this and zpk and ly have firstPath: "path G ?pk y" by simp
	from B have "\<exists> i \<ge> l+k. x = (p i, p (Suc i))" ..
	from this obtain i where ilk: "i \<ge> l + k" and xi: "x = (p i, p (Suc i))" by auto
	from ilk have "\<exists> j. i =  l + j + k" by arith
	from this obtain j where iklj: "i = l + j + k" ..
	from this have tmp1: "?q (l+j) = p i" by (unfold drop_path_def, simp)
	from qPath have "path G (?q l) (?q (l+j))" by (rule is_path_imp_path, simp)
	from this and tmp1 and xi and ly and xpk have "path G y ?pk" by simp
	from firstPath and this show "y \<in> scc_of_node G ?pk" by (rule cycle_scc)
     qed
     from D and C and this show ?thesis by auto 
   qed
qed

lemma path_nodes : assumes "is_path G p" and "path_nodes p \<subseteq> C" shows "is_path (G|\<^sub>\<^sub>\<^sub>C) p" using assms
proof (unfold is_path_def path_nodes_def node_restriction_def edges_def, simp)
  assume pC: "range p \<subseteq> C"
  show "\<forall>n. p n \<in> C \<and> p (Suc n) \<in> C"
  proof 
    fix n
    from pC have nC: "p n \<in> C" by auto
    from pC have "p (Suc n) \<in> C" by auto
    from nC and this show "p n \<in> C \<and> p (Suc n) \<in> C" ..
  qed
qed

definition real_scc :: "'a set \<Rightarrow> 'a graph \<Rightarrow> bool"
where "real_scc C G = (scc C G \<and> (\<forall> x. C = {x} \<longrightarrow> (x, x) \<in> edges G))"

lemma path_ends_in_scc : assumes "finite (edges G)" and "is_path G p" 
  shows "\<exists> q C. real_scc C G \<and> p =\<dots> q \<and> is_path (G|\<^sub>\<^sub>\<^sub>C) q"  
using assms
proof -
  from `is_path G p` have "path_edges p \<subseteq> edges G" by (unfold path_edges_def is_path_def, blast)
  from `finite (edges G)` and `is_path G p` and this have "\<exists> q C. scc C G \<and> p =\<dots> q \<and> path_nodes q \<subseteq> C" 
  by (rule main_scc_helper)
  from this obtain q C where A: "scc C G" and B: "p =\<dots> q" and C: "path_nodes q \<subseteq> C" by auto
  from `is_path G p` and B have "is_path G q" by (rule is_path_imp_suffix_is_path)
  from this and C have D: "is_path (G|\<^sub>\<^sub>\<^sub>C) q" by (rule path_nodes) 
  have "\<forall> x. C = {x} \<longrightarrow> (x, x) \<in> edges G" using D 
  proof -
    assume E: "is_path (G|\<^sub>\<^sub>\<^sub>C) q"
    show ?thesis 
    proof (rule allI, rule impI)
      fix x
      assume Cx: "C = {x}"
      from E have "\<forall>n. (q n, q (Suc n)) \<in> edges (C, {(m, n). m \<in> C \<and> n \<in> C \<and> (m, n) \<in> edges G})"
	by (unfold is_path_def node_restriction_def edges_def, simp)
      hence "(q 0, q 1) \<in> {(m, n). m \<in> C \<and> n \<in> C \<and> (m,n) \<in> edges G}" by (unfold edges_def,simp)
      hence "q 0 \<in> C" and "q 1 \<in> C" and "(q 0, q 1) \<in> edges G" by auto
      from this and `C = {x}` show "(x,x) \<in> edges G" by simp
    qed
  qed
  from A and this have "real_scc C G" by (unfold real_scc_def, simp)
  from this and B and D show ?thesis by auto
qed


(* start of automated check *)



definition graph :: "'a list \<Rightarrow> ('a \<times> 'a) list \<Rightarrow> 'a graph"
where "graph n e = (set n, set e)" 

definition graph2 :: "'a list \<Rightarrow> ('a \<times> 'a) set \<Rightarrow> 'a graph"
where "graph2 n e = (set n, e)"

fun check_edge :: "('a \<times> 'a) \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> bool"
where "check_edge e [] = False"
   |  "check_edge (x,y) ((real,C) # cs) = ((x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda> D. y mem (snd D)) (cs))) \<or> check_edge (x,y) cs)"


lemma check_edge_sound : assumes "check_edge (x,y) cs" shows "\<exists> i j. i \<le> j \<and> j < length cs \<and> x mem (snd (cs ! i)) \<and> y mem (snd (cs ! j)) \<and> (i = j \<longrightarrow> fst (cs ! i))" using assms
proof (induct cs)
  case Nil
  thus ?case by simp
next
  case (Cons rC cs)
  show ?case
  proof (cases rC)
    case (Pair real C)
    with `check_edge (x,y) (rC # cs)` have " (x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda>D. y mem (snd D)) cs)) \<or> check_edge (x,y) cs" by simp
    thus ?thesis
    proof 
      assume "check_edge (x,y) cs"
      with Cons obtain i j where "i \<le> j" and "j < length cs" and "x mem (snd  (cs ! i))" and "y mem (snd (cs ! j))" and "i = j \<longrightarrow> fst (cs ! i)" by auto
      hence "Suc i \<le> Suc j \<and> Suc j < length (rC # cs) \<and> x mem (snd ((rC # cs) ! (Suc i))) \<and> y mem (snd ((rC # cs) ! (Suc j))) \<and> (Suc i = Suc j \<longrightarrow> fst ((rC # cs) ! (Suc i)))" by auto
      thus ?thesis by blast
    next
      assume Ass: "x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda> D . y mem (snd D)) cs)" (is "_ \<and> (?one \<or> ?two)")
      hence xC: "x mem C" and "?one \<or> ?two" by auto
      from xC Pair have x: "\<forall> j. (x mem (snd ((rC # cs) ! 0)) \<and> 0 \<le> j \<and> 0 \<le> Suc j \<and> 0 < length (rC # cs))"  by auto
      from `?one \<or> ?two` 
      show ?thesis
      proof
	assume ?two
	hence "\<exists> j < length cs. y mem (snd (cs ! j))" by (simp add: testPosSound2)
	from this obtain j where "y mem (snd (cs ! j))" and "j < length cs" by auto
	hence  "y mem (snd ((rC # cs) ! (Suc j)))" and "Suc j < length (rC # cs)" and "0 < Suc j" by auto
	with x show ?thesis by blast
      next
	assume ?one
	with Pair have "y mem (snd ((rC # cs) ! 0)) \<and> 0 < length (rC # cs) \<and> fst ((rC # cs) ! 0)"  by auto
	with x Pair show ?thesis by blast
      qed
    qed
  qed
qed


(* the idea is to first look whether the edge is allowed due to decomposition and only 
  otherwise really check whether the edge is in the graph *)
fun check_edge_lazy :: "(('a \<times> 'a) \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'a) \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> bool"
where "check_edge_lazy g e [] = (\<not> g e)"
   |  "check_edge_lazy g (x,y) ((real,C) # cs) = ((x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda> D. y mem (snd D)) (cs))) \<or> check_edge_lazy g (x,y) cs)"

lemma check_edge_lazy_sound : assumes check: "check_edge_lazy g (x,y) cs" and xyg: "(x,y) \<in> g" 
  shows "\<exists> i j. i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)"
  using check
proof (induct cs)
  case Nil
  with xyg show ?case by (simp add: mem_def)
next
  case (Cons rC cs)
  show ?case
  proof (cases rC)
    case (Pair real C)
    with `check_edge_lazy g (x,y) (rC # cs)` have "(x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda>D. y mem (snd D)) cs)) \<or> check_edge_lazy g (x,y) cs" by simp
    thus ?thesis
    proof 
      assume "check_edge_lazy g (x,y) cs"
      with Cons obtain i j where "i \<le> j" and "j < length cs" and "x \<in> set (snd  (cs ! i))" and "y \<in> set (snd (cs ! j))" and "(\<not> fst (cs ! i) \<longrightarrow> i < j)" by auto
      hence "Suc i \<le> Suc j \<and> Suc j < length (rC # cs) \<and> x \<in> set (snd ((rC # cs) ! (Suc i))) \<and> y \<in> set (snd ((rC # cs) ! (Suc j))) \<and> ((\<not> fst ((rC # cs) ! (Suc i)) \<longrightarrow> (Suc i) < (Suc j)))" by auto
      thus ?thesis by blast
    next
      assume Ass: "x mem C \<and> ((real \<and> y mem C) \<or> testPos (\<lambda> D . y mem (snd D)) cs)" (is "_ \<and> (?one \<or> ?two)")
      hence xC: "x \<in> set C" and "?one \<or> ?two" by (auto simp: mem_iff)
      from xC Pair have x: "\<forall> j. (x \<in> set (snd ((rC # cs) ! 0)) \<and> 0 \<le> j \<and> 0 \<le> Suc j \<and> 0 < length (rC # cs))"  by auto
      from `?one \<or> ?two` 
      show ?thesis
      proof
	assume ?two
	hence "\<exists> j < length cs. y mem (snd (cs ! j))" by (simp add: testPosSound2)
	from this obtain j where "y \<in> set (snd (cs ! j))" and "j < length cs" by (auto simp: mem_iff)
	hence  "y \<in> set (snd ((rC # cs) ! (Suc j)))" and "Suc j < length (rC # cs)" and "0 < Suc j" by auto
	with x show ?thesis by blast
      next
	assume ?one
	with Pair have "y \<in> set (snd ((rC # cs) ! 0)) \<and> 0 < length (rC # cs) \<and> fst ((rC # cs) ! 0)"  by (auto simp: mem_iff)
	with x Pair show ?thesis by blast
      qed
    qed
  qed
qed

definition check_edgeM :: "(('a \<times> 'a) \<Rightarrow> bool) \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> ('a \<times> 'a) \<Rightarrow> ('a \<times> 'a) monad" where "check_edgeM g cs e \<equiv> check (\<lambda> e. check_edge_lazy g e cs) e"

lemma check_edgeMSound[simp] : "\<lbrakk>isOK (check_edgeM g cs (x,y))\<rbrakk> \<Longrightarrow> (x,y) \<in> g \<longrightarrow> (\<exists> i j. i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j))"
using assms unfolding check_edgeM_def
by (simp only: checkFSound, rule impI, auto simp: check_edge_lazy_sound)



definition checkSccLazy :: "('a  \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> ('a \<times> 'a)monad"
where "checkSccLazy g ns cs \<equiv> checkAllM (check_edgeM g cs) [(x,y). x \<leftarrow> ns, y \<leftarrow> ns]"


lemma checkSccLazySound[simp] : assumes check: "isOK(checkSccLazy g n cs)"
                        shows   "\<forall> (x,y) \<in> edges (graph2 n ((set n \<times> set n) \<inter> g)). \<exists> i j . i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" 
proof - 
  show ?thesis unfolding edges_def graph2_def 
  proof (auto simp:mem_iff)
    fix x y
    assume xyg: "(x,y) \<in> g" and "x \<in> set n" and "y \<in> set n"
    with check have "isOK(check_edgeM g cs (x, y))"
      unfolding checkSccLazy_def by auto
    with xyg show "\<exists> i j . i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" by auto 
  qed
qed


abbreviation graph3  where "graph3 g n \<equiv> graph2 n ((set n \<times> set n) \<inter> g)"

lemma finiteGraph : "finite (edges (graph3 g n))"
  by (unfold graph2_def edges_def, auto)
  
  
primrec index :: "('a set) list \<Rightarrow> 'a \<Rightarrow> nat"
where "index [] x = 0"
  |   "index (C # cs) x = (if x \<in> C then 0 else Suc (index cs x))"


lemma uniqueIndex : assumes "partition cs" 
                        and "x \<in> cs ! i"
                        and "i < length cs"
                      shows "index cs x = i"
using assms
proof (induct cs arbitrary: i)
  case Nil thus ?case by simp
next
  case (Cons C cs)
  show ?case
  proof (cases i)
    case 0
    with Cons show ?thesis by simp
  next
    case (Suc j)    
    with Cons have jl: "j < length cs" by simp
    have di: "partition cs"
    proof (unfold partition_def, (rule allI impI)+)
      fix j i
      assume "j < length cs" and "i < j"
      hence lj: "Suc j < length (C # cs)" and li: "Suc i < Suc j" by auto
      hence "(C # cs) ! (Suc i) \<inter> (C # cs) ! (Suc j) = cs ! i \<inter> cs ! j" by auto
      with `partition (C # cs)` and lj and li show "cs ! i \<inter> cs ! j = {}" by (unfold partition_def, blast)
    qed
    from Suc and Cons have "x \<in> cs ! j" by simp
    with jl and di and Cons have indexJ: "index cs x = j" by simp
    have "0 < length (C # cs)" by simp
    with `partition (C # cs)` and Suc and `i < length (C # cs)`  have "(C # cs) ! 0 \<inter> (C # cs) ! i = {}" by (unfold partition_def, blast)
    with `x \<in> (C # cs) ! i` have "x \<notin> C" by auto
    with indexJ and Suc show ?thesis by simp
  qed
qed

lemma checkSccSoundIII : 
   assumes fP: "is_fin_path (graph3 g n) xs" (is "is_fin_path ?G xs")
         and  ch: "isOK(checkSccLazy g n cs)"
         and  di: "partition (map set (map snd cs))" (is "partition ?cs")
        shows "sorted (map (index (map set (map snd cs))) xs)"
     using fP
proof (induct xs)
  case Nil thus ?case by simp
next
  case (Cons y ys)
  hence IH1: "is_fin_path ?G ys \<Longrightarrow> sorted (map (index ?cs) ys)" 
    and IH2: "is_fin_path ?G (y # ys)" by auto
  thus ?case
  proof (cases ys)
    case Nil thus ?thesis by simp
  next
    case (Cons z zs)
    with IH2 and IH1 have "sorted (map (index ?cs) ys)" by simp
    with Cons show ?thesis
    proof (simp)
      show "index ?cs y \<le> index ?cs z"
      proof -
	from IH2 and Cons have yz: "(y,z) \<in> edges ?G" by simp
	from ch have "\<forall> (y,z) \<in> edges ?G. \<exists> i j . i \<le> j \<and> j < length cs \<and> y \<in> set (snd (cs ! i)) \<and> z \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" 
	  by simp
	with yz have "\<exists> i j . i \<le> j \<and> j < length cs \<and> y \<in> set (snd (cs ! i)) \<and> z \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" by auto 
	from this obtain i j where ij: "i \<le> j" and jl: "j < length ?cs" and yi: "y \<in> ?cs ! i" and zj: "z \<in> ?cs ! j" by auto
	hence il: "i < length ?cs" by simp
	from di and yi and il have iy: "index ?cs y = i" by (rule uniqueIndex) 	
	from di and zj and jl have "index ?cs z = j" by (unfold partition_def, simp add: uniqueIndex) 
	with iy and ij show ?thesis by simp
      qed
    qed
  qed
qed


lemma sortHelp: assumes "sorted (map f xs)" 
        and "i \<le> j"
        and "j < length xs"
      shows "f (xs ! i) \<le> f (xs ! j)"
using assms 
proof (induct xs arbitrary: i j)
  case Nil thus ?case by simp
next
  case (Cons y ys)
  from Cons have sort: "sorted (map f ys)" by (cases ys, auto)
  show ?case 
  proof (cases i)
    case (Suc k)
    with `i \<le> j` have "\<exists> l. j = Suc l" by arith
    from this obtain l where "j = Suc l" ..
    with Cons and Suc and sort show ?thesis by auto
  next
    case 0
    show ?thesis 
    proof (cases "i = j")
      case True
      with 0 and Cons show ?thesis by simp
    next
      case False
      with 0 have "\<exists> l. j = Suc l" by arith
      from this obtain l where jl: "j = Suc l" ..
      have "0 \<le> l" by simp
      with Cons and sort and jl have "f (ys ! 0) \<le> f (ys ! l)" by simp
      with Cons and jl have step: "f (ys ! 0) \<le> f ((y # ys) ! j)" by simp 
      from `sorted (map f (y # ys))` and jl and `j < length (y # ys)` have "f y \<le> f (ys ! 0)" by (cases ys, auto)
      with step and Cons and 0 show ?thesis by simp
    qed
  qed
qed

lemma checkSccSoundIV : 
   assumes fp: "is_fin_path (graph3 g n) (x#xs)" (is "is_fin_path ?G _")
  and  ch: "isOK(checkSccLazy g n cs)"
  and  di: "partition (map set (map snd cs))" (is "partition ?cs")
  shows "index (map set (map snd cs)) x \<le> index (map set (map snd cs)) (last (x # xs))"
proof -
  from fp and ch and di have sort: "sorted (map (index ?cs) (x#xs))" by (rule checkSccSoundIII)
  let ?n = "length (x # xs) - 1"
  have "0 \<le> ?n" and "?n < length (x # xs)" by auto
  from sort and this have ineq: "index ?cs ((x # xs) ! 0) \<le> index ?cs ((x # xs) ! ?n)" by (rule sortHelp)
  have x0: "x = (x # xs) ! 0" by simp
  have "x # xs \<noteq> []" by simp
  hence "last (x # xs) = (x # xs) ! ?n" by (rule last_nth)
  with x0 and ineq show ?thesis by simp
qed
  
lemma sorted_equal :
  assumes A: "sorted (map (index cs) xs)" 
       and B: "index cs (hd xs) = index cs (last xs)"
       and C: "x \<in> set xs"
  shows "index cs x = index cs (hd xs)"
using assms 
proof -
  from C have  "\<exists> i . xs ! i = x \<and> i < length xs" by (rule list_set)
  from this obtain i where D: "xs ! i = x" and E: "i < length xs" by auto
  from E have nE: "xs \<noteq> []" by auto
  have "0 \<le> i" by arith
  from A and this and E have ineq1: "index cs (xs ! 0) \<le> index cs (xs ! i)" by (rule sortHelp)
  let ?n = "length xs - 1"
  from E have "i \<le> ?n" and "?n < length xs" by auto
  from A and this have ineq2: "index cs (xs ! i) \<le> index cs (xs ! ?n)" by (rule sortHelp)
  from nE have eq1: "xs ! 0 = hd xs" by (cases xs, auto)
  from nE have eq2: "last xs = xs ! ?n" by (rule last_nth)
  from ineq1 and ineq2 and eq1 and eq2 and B and D show ?thesis by simp
qed
  

lemma checkSccSoundV :
  assumes   A: "isOK (checkSccLazy g n cs)"
     and    B: "partition (map set (map snd cs))" (is "partition ?cs") 
     and    C: "real_scc S (graph3 g n)" (is "real_scc S ?G")
      shows "\<exists> C \<in> set (map snd (filter fst cs)). set C \<supseteq> S"
proof -
  from C have "cycle S ?G" by (unfold real_scc_def scc_def, simp)
  hence sNE: "S \<noteq> {}" and sPaths: "\<forall>m\<in>S. \<forall>n\<in>S. path (?G|\<^sub>\<^sub>\<^sub>S) m n" by (unfold cycle_def, auto)
  have Si : "\<forall> x \<in> S. x \<in> ?cs ! (index ?cs x) \<and> index ?cs x < length cs \<and> (S \<subseteq> ?cs ! (index ?cs x) \<longrightarrow> fst (cs ! (index ?cs x)))"
  proof
    fix x
    assume xS: "x \<in> S"
    have "\<exists> y \<in> S . (x,y) \<in> edges ?G"
    proof (cases "S = {x}")
      case True
      with C have "(x,x) \<in> edges ?G" by (unfold real_scc_def, simp)
      with xS show ?thesis by blast
    next
      case False
      from this and xS have "\<exists> z \<in> S. z \<noteq> x" by auto
      from this obtain z where zS: "z \<in> S" and xz: "z \<noteq> x" by auto
      from xS and zS and sPaths have "path (?G|\<^sub>\<^sub>\<^sub>S) x z" by simp
      thus ?thesis
      proof (cases rule: path.cases)
	case refl
	with xz show ?thesis by simp
      next
	case (trans u y g)
	from trans have edge: "(x,y) \<in> edges ?G" by (unfold node_restriction_def edges_def, auto) 
	from trans have "y \<in> S" by (unfold node_restriction_def edges_def, auto)
	with edge show ?thesis by blast
      qed
    qed
    from this obtain y where yS: "y \<in> S" and edge: "(x,y) \<in> edges ?G" by auto
    from A have contain: " \<forall>(x, y)\<in>edges ?G. \<exists>i j. i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" by simp
    with edge have "\<exists>i j. i \<le> j \<and> j < length cs \<and> x \<in> set (snd (cs ! i)) \<and> y \<in> set (snd (cs ! j)) \<and> (\<not> fst (cs ! i) \<longrightarrow> i < j)" by blast
    from this obtain i j where "i \<le> j" and jl: "j < length ?cs" and xi: "x \<in> ?cs ! i" and yj: "y \<in> ?cs ! j"
      and fst: "\<not> fst (cs ! i) \<longrightarrow> i < j" by auto
    hence il: "i < length ?cs" by simp
    from B and xi and this have ixi: "index ?cs x = i" by (rule uniqueIndex)
    from xi and il and ixi 
    have xind: "x \<in> ?cs ! index ?cs x \<and> index ?cs x < length cs" by auto
    hence xind2: "index ?cs x < length ?cs" by simp
    from B and yj and jl have iyj: "index ?cs y = j" by (rule uniqueIndex)
    { 
      assume "S \<subseteq> ?cs ! index ?cs x"
      with yS have "y \<in> ?cs ! index ?cs x" by auto
      from B and this and xind2 have "index ?cs y = index ?cs x" by (rule uniqueIndex)
      with ixi iyj have "i = j" by simp
      with fst have "fst (cs ! i)" by simp
      with ixi have "fst (cs ! index ?cs x)" by simp
    }
    with xind 
    show "x \<in> ?cs ! index ?cs x \<and> index ?cs x < length cs \<and> 
      (S \<subseteq> ?cs ! index ?cs x \<longrightarrow> fst (cs ! (index ?cs x)))" by auto
  qed

  from sNE have "\<exists> x. x \<in> S" by auto
  from this obtain x where xS: "x \<in> S" ..  
  let ?ix = "index ?cs x"
  let ?C = "?cs ! ?ix" (* we have found the component containing x *)
  from xS and Si have xC: "x \<in> ?C" and IXL: "?ix < length cs" by auto

  have CS: "?C \<supseteq> S" 
  proof
    fix y
    let ?iy = "index ?cs y"
    let ?D = "?cs ! ?iy"
    assume yS: "y \<in> S"    
    from this and Si have yD: "y \<in> ?D" by auto
    show "y \<in> ?C"
    proof -
      from xS and yS and sPaths have "path (?G|\<^sub>\<^sub>\<^sub>S) x y" by simp
      hence "\<exists> xs. last (x # xs) = y \<and> is_fin_path (?G|\<^sub>\<^sub>\<^sub>S) (x # xs)" by (rule path_exists)
      from this obtain xs where lastXY: "last (x # xs) = y" and pXY: "is_fin_path (?G|\<^sub>\<^sub>\<^sub>S) (x # xs)" by auto
      from pXY have "is_fin_path ?G (x # xs)" by (rule is_fin_path_restriction)
      from this and A and B have xIy: "?ix \<le> index ?cs (last (x # xs))" by (rule checkSccSoundIV)
      from xS and yS and sPaths have "path (?G|\<^sub>\<^sub>\<^sub>S) y x" by simp
      hence "\<exists> ys. last (y # ys) = x \<and> is_fin_path (?G|\<^sub>\<^sub>\<^sub>S) (y # ys)" by (rule path_exists)
      from this obtain ys where lastYX: "last (y # ys) = x" and pYX: "is_fin_path (?G|\<^sub>\<^sub>\<^sub>S) (y # ys)" by auto
      from pYX have "is_fin_path ?G (y # ys)" by (rule is_fin_path_restriction) 
      from this and A and B have yIx: "?iy \<le> index ?cs (last (y # ys))" by (rule checkSccSoundIV)
      from xIy and yIx and lastXY and lastYX have "?ix = ?iy" by simp
      with yD and xC show ?thesis by simp
    qed
  qed

  with Si xS have fst: "fst (cs ! ?ix)" by auto
  have "\<forall> f g i. i < length cs
    \<and> g (cs ! i) \<longrightarrow> map set (map f cs) ! i \<in> set (map set (map f (filter g cs)))" 
    by auto 
  with fst IXL have "?C \<in> set (map set (map snd (filter fst cs)))" by blast

  with CS show ?thesis by force
qed


lemma checkSccSoundVI :
  assumes   A: "isOK (checkSccLazy g n cs) "
     and    B: "isOK(checkPartition (map snd cs))"
     and    C: "is_path (graph3 g n) p" (is "is_path ?G p")
   shows "\<exists> q C . p =\<dots> q \<and> C \<in> set (map snd (filter fst cs)) \<and> is_path (?G|\<^sub>\<^sub>\<^sub>(set C)) q"
proof -
  from B have B2: "partition (map set (map snd cs))" by simp
  have "finite (edges ?G)" by (rule finiteGraph)
  from this and C have "\<exists>q S. real_scc S ?G \<and> p =\<dots> q \<and> is_path (?G|\<^sub>\<^sub>\<^sub>S) q" by (rule path_ends_in_scc)
  from this obtain q S where rS: "real_scc S ?G" and suff: "p =\<dots> q" and path: "is_path (?G|\<^sub>\<^sub>\<^sub>S) q" by auto
  from A and B2 and rS have "\<exists> C \<in> set (map snd (filter fst cs)). S \<subseteq> set C" by (rule checkSccSoundV)
  from this obtain C where cs: "C \<in> set (map snd (filter fst cs))" and subs: "S \<subseteq> set C" by blast
  from path and subs have "is_path (?G|\<^sub>\<^sub>\<^sub>(set C)) q" by (rule is_path_monotone)
  with cs and suff  show ?thesis by blast
qed

(* 
   edge list: the list of edges that represent the graph

   allSccs: has to be a list of sccs in order of edges. I.e., if x \<rightarrow> y then scc of x must appear before y.
     Moreover, allSccs has to contain singleton nodes which don't have an edge to itself
     However, these should be marked with False in the first component.


   if these conditions are met, then checkPartition will deliver True (otherwise there would be a bug)
*)


definition checkDecomposition :: "('a \<times> 'a :: sshow) set \<Rightarrow> 'a list \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> sstring list monad"
where "checkDecomposition g n allSccs \<equiv> 
  checkPartition (map snd allSccs) \<leftarrow>
  (\<lambda> e. [''components do not form a partition due to node '' +++ sshow e]) \<guillemotright>
  checkSccLazy g n allSccs \<leftarrow> 
  (\<lambda> e. [''the following edge is not covered:'',sshow e])
"


lemma checkDecompositionSound :
      assumes A: "isOK(checkDecomposition g n cs)"
       and    B: "is_path (graph2 n g) p"
       and    C: "path_nodes p \<subseteq> set n" 
      shows "\<exists> q C . p =\<dots> q \<and> C \<in> set (map snd (filter fst cs)) \<and> path_nodes q \<subseteq>  set C" 
proof -
  let ?G = "graph3 g n"
  fix es ns
  from B and C have B2: "is_path ?G p" unfolding path_nodes_def 
     graph2_def is_path_def edges_def by auto
  from A have "isOK (checkSccLazy g n cs)"  and 
    "isOK(checkPartition (map snd cs))"
    unfolding checkDecomposition_def by auto
  from this and B2 have "\<exists> q C . p =\<dots> q \<and> C \<in> set (map snd (filter fst cs)) \<and> is_path (?G|\<^sub>\<^sub>\<^sub>(set C)) q" by (rule checkSccSoundVI) 
  from this obtain q C where one: "p =\<dots> q" and two: "C \<in> set (map snd (filter fst cs))" and three: "is_path (?G|\<^sub>\<^sub>\<^sub>(set C)) q" by auto
  from three have four: "path_nodes q \<subseteq> set C"  unfolding 
    node_restriction_def nodes_def edges_def is_path_def path_nodes_def 
  by auto
  with one two show ?thesis by force 
qed




end
